from __future__ import annotations

import os
from typing import Iterator, List, Optional, Sequence

from ...canonical_models import JiraIssue
from ...errors import SerializationError
from ..client import JiraRestClient
from ..env import auth_from_env, jira_rest_base_url_from_env
from ..gen import jira_api as api
from ..mappers.jira_issues import map_issue


_DEFAULT_SEARCH_FIELDS = (
    "project",
    "issuetype",
    "status",
    "created",
    "updated",
    "resolutiondate",
    "assignee",
    "reporter",
    "labels",
    "components",
)


def _build_field_list(
    fields: Optional[Sequence[str]],
    story_points_field: Optional[str],
    sprint_ids_field: Optional[str],
) -> List[str]:
    field_list = [
        f.strip() for f in (fields or _DEFAULT_SEARCH_FIELDS) if f and f.strip()
    ]
    if not field_list:
        raise ValueError("fields must be non-empty")
    for raw in (story_points_field, sprint_ids_field):
        if raw is None:
            continue
        extra = raw.strip()
        if not extra:
            raise ValueError("custom field names must be non-empty when provided")
        if extra not in field_list:
            field_list.append(extra)
    return field_list


def _env_custom_field(env_name: str) -> Optional[str]:
    raw = os.getenv(env_name)
    if raw is None:
        return None
    cleaned = raw.strip()
    return cleaned or None


def iter_issues_via_rest(
    client: JiraRestClient,
    cloud_id: str,
    jql: str,
    page_size: int = 50,
    *,
    fields: Optional[Sequence[str]] = None,
    story_points_field: Optional[str] = None,
    sprint_ids_field: Optional[str] = None,
) -> Iterator[JiraIssue]:
    cloud_id_clean = (cloud_id or "").strip()
    if not cloud_id_clean:
        raise ValueError("cloud_id is required")
    jql_clean = (jql or "").strip()
    if not jql_clean:
        raise ValueError("jql is required")
    if page_size <= 0:
        raise ValueError("page_size must be > 0")

    if story_points_field is None:
        story_points_field = _env_custom_field("ATLASSIAN_JIRA_STORY_POINTS_FIELD")
    if sprint_ids_field is None:
        sprint_ids_field = _env_custom_field("ATLASSIAN_JIRA_SPRINT_IDS_FIELD")

    field_list = _build_field_list(fields, story_points_field, sprint_ids_field)
    fields_param = ",".join(field_list)

    start_at = 0
    seen_start_at: set[int] = set()

    while True:
        if start_at in seen_start_at:
            raise SerializationError(
                "Pagination startAt repeated; aborting to prevent infinite loop"
            )
        seen_start_at.add(start_at)

        payload = client.get_json(
            "/rest/api/3/search",
            params={
                "jql": jql_clean,
                "startAt": start_at,
                "maxResults": page_size,
                "fields": fields_param,
            },
        )
        page = api.SearchResults.from_dict(payload, "data")
        issues = page.issues

        for issue in issues:
            yield map_issue(
                cloud_id=cloud_id_clean,
                issue=issue,
                story_points_field=story_points_field,
                sprint_ids_field=sprint_ids_field,
            )

        has_total = isinstance(page.total, int) and page.total >= 0
        if has_total:
            if start_at + len(issues) >= page.total:
                break
        else:
            if len(issues) < page_size:
                break

        if len(issues) == 0:
            break
        start_at += len(issues)


def list_issues_via_rest(
    cloud_id: str,
    jql: str,
    page_size: int = 50,
    *,
    fields: Optional[Sequence[str]] = None,
    story_points_field: Optional[str] = None,
    sprint_ids_field: Optional[str] = None,
) -> Iterator[JiraIssue]:
    cloud_id_clean = (cloud_id or "").strip()
    if not cloud_id_clean:
        raise ValueError("cloud_id is required")

    auth = auth_from_env()
    if auth is None:
        raise ValueError(
            "Missing credentials. Set ATLASSIAN_OAUTH_ACCESS_TOKEN, or "
            "ATLASSIAN_OAUTH_REFRESH_TOKEN + (ATLASSIAN_CLIENT_ID + ATLASSIAN_CLIENT_SECRET), "
            "or (ATLASSIAN_EMAIL + ATLASSIAN_API_TOKEN), or ATLASSIAN_COOKIES_JSON."
        )

    base_url = jira_rest_base_url_from_env(cloud_id_clean)
    if not base_url:
        raise ValueError(
            "Missing Jira REST base URL. Set ATLASSIAN_JIRA_BASE_URL, or "
            "provide ATLASSIAN_CLOUD_ID with OAuth tokens (defaults to https://api.atlassian.com/ex/jira/{cloudId}), "
            "or provide ATLASSIAN_GQL_BASE_URL for tenanted auth so it can be derived."
        )

    if story_points_field is None:
        story_points_field = _env_custom_field("ATLASSIAN_JIRA_STORY_POINTS_FIELD")
    if sprint_ids_field is None:
        sprint_ids_field = _env_custom_field("ATLASSIAN_JIRA_SPRINT_IDS_FIELD")

    with JiraRestClient(
        base_url, auth=auth, timeout_seconds=30.0, max_retries_429=1
    ) as client:
        yield from iter_issues_via_rest(
            client,
            cloud_id=cloud_id_clean,
            jql=jql,
            page_size=page_size,
            fields=fields,
            story_points_field=story_points_field,
            sprint_ids_field=sprint_ids_field,
        )


def create_issue_via_rest(
    cloud_id: str,
    project_key: str,
    summary: str,
    issue_type: str = "Task",
    description: Optional[str] = None,
) -> JiraIssue:
    auth = auth_from_env()
    if auth is None:
        raise ValueError("Missing credentials.")

    base_url = jira_rest_base_url_from_env(cloud_id)
    if not base_url:
        raise ValueError("Missing Jira REST base URL.")

    with JiraRestClient(base_url, auth=auth) as client:
        data = {
            "fields": {
                "project": {"key": project_key},
                "summary": summary,
                "issuetype": {"name": issue_type},
            }
        }
        if description:
            data["fields"]["description"] = description

        payload = client.post_json("/rest/api/3/issue", json_data=data)
        issue_bean = api.IssueBean.from_dict(payload, "data")
        return map_issue(cloud_id=cloud_id, issue=issue_bean)


def delete_issue_via_rest(
    cloud_id: str,
    issue_key_or_id: str,
) -> None:
    auth = auth_from_env()
    if auth is None:
        raise ValueError("Missing credentials.")

    base_url = jira_rest_base_url_from_env(cloud_id)
    if not base_url:
        raise ValueError("Missing Jira REST base URL.")

    with JiraRestClient(base_url, auth=auth) as client:
        client.delete(f"/rest/api/3/issue/{issue_key_or_id}")



from __future__ import annotations

import logging
from typing import Dict, Mapping


def get_logger(logger: logging.Logger | None = None) -> logging.Logger:
    return logger if logger is not None else logging.getLogger("atlassian")


def sanitize_headers(headers: Mapping[str, str]) -> Dict[str, str]:
    sensitive = {"authorization", "cookie"}
    sanitized: Dict[str, str] = {}
    for key, value in headers.items():
        if key.lower() in sensitive:
            sanitized[key] = "<redacted>"
        else:
            sanitized[key] = value
    return sanitized



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional


@dataclass
class GraphQLErrorItem:
    message: str
    path: Optional[List[Any]] = None
    extensions: Optional[Dict[str, Any]] = None
    locations: Optional[List[Dict[str, Any]]] = None


@dataclass
class GraphQLResult:
    data: Optional[Dict[str, Any]]
    errors: Optional[List[GraphQLErrorItem]]
    extensions: Optional[Dict[str, Any]]


def parse_error_items(raw_errors: Any) -> Optional[List[GraphQLErrorItem]]:
    if raw_errors is None:
        return None
    items: List[GraphQLErrorItem] = []
    if not isinstance(raw_errors, list):
        return None
    for err in raw_errors:
        if not isinstance(err, dict):
            continue
        message = err.get("message")
        if not isinstance(message, str):
            continue
        items.append(
            GraphQLErrorItem(
                message=message,
                path=err.get("path"),
                extensions=err.get("extensions"),
                locations=err.get("locations"),
            )
        )
    return items



import httpx

from atlassian.auth import OAuthBearerAuth
from atlassian.graph.api.jira_projects import iter_projects_with_opsgenie_linkable_teams
from atlassian.graph.client import GraphQLClient


def test_beta_headers_sent_multiple_times():
    captured = {}

    def capture(request: httpx.Request):
        captured["beta"] = request.headers.get_list("X-ExperimentalApi")
        return httpx.Response(200, json={"data": {"ok": True}}, request=request)

    transport = httpx.MockTransport(capture)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://beta.example.com",
            auth=OAuthBearerAuth(lambda: "token"),
            http_client=http_client,
        )
        client.execute("query { ok }", experimental_apis=["featureA", "featureB"])
        assert captured["beta"] == ["featureA", "featureB"]


def test_beta_headers_sent_on_jira_projects_requests():
    betas: list[list[str]] = []
    call_count = {"count": 0}

    def handler(request: httpx.Request):
        betas.append(request.headers.get_list("X-ExperimentalApi"))
        call_count["count"] += 1
        return httpx.Response(
            200,
            json={
                "data": {
                    "jira": {
                        "projects": {
                            "pageInfo": {"hasNextPage": False, "endCursor": None},
                            "edges": [],
                        }
                    }
                }
            },
            request=request,
        )

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://beta.example.com",
            auth=OAuthBearerAuth(lambda: "token"),
            http_client=http_client,
        )
        list(
            iter_projects_with_opsgenie_linkable_teams(
                client,
                cloud_id="cloud-123",
                project_types=["SOFTWARE"],
                page_size=10,
                experimental_apis=["featureA", "featureB"],
            )
        )

    assert call_count["count"] == 1
    assert betas and betas[0] == ["featureA", "featureB"]



import httpx

from atlassian.auth import OAuthBearerAuth
from atlassian.rest.api.jira_worklogs import iter_issue_worklogs_via_rest
from atlassian.rest.client import JiraRestClient


def test_jira_rest_worklogs_pagination_and_mapping():
    def handler(request: httpx.Request) -> httpx.Response:
        assert request.url.path.endswith("/rest/api/3/issue/A-1/worklog")
        start_at = int(request.url.params.get("startAt", "0"))
        if start_at == 0:
            return httpx.Response(
                200,
                json={
                    "startAt": 0,
                    "maxResults": 1,
                    "total": 2,
                    "worklogs": [
                        {
                            "id": "200",
                            "author": {"accountId": "u1", "displayName": "User 1"},
                            "started": "2021-01-02T00:00:00.000+0000",
                            "timeSpentSeconds": 60,
                            "created": "2021-01-02T00:00:00.000+0000",
                            "updated": "2021-01-02T00:00:00.000+0000",
                        }
                    ],
                },
                request=request,
            )
        if start_at == 1:
            return httpx.Response(
                200,
                json={
                    "startAt": 1,
                    "maxResults": 1,
                    "total": 2,
                    "worklogs": [
                        {
                            "id": "201",
                            "started": "2021-01-03T00:00:00.000+0000",
                            "timeSpentSeconds": 120,
                            "created": "2021-01-03T00:00:00.000+0000",
                            "updated": "2021-01-03T00:00:00.000+0000",
                        }
                    ],
                },
                request=request,
            )
        raise AssertionError(f"unexpected startAt={start_at}")

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = JiraRestClient(
            "https://api.atlassian.com/ex/jira/cloud-123",
            auth=OAuthBearerAuth(lambda: "token"),
            http_client=http_client,
        )
        worklogs = list(iter_issue_worklogs_via_rest(client, issue_key="A-1", page_size=1))

    assert len(worklogs) == 2
    assert worklogs[0].issue_key == "A-1"
    assert worklogs[0].worklog_id == "200"
    assert worklogs[0].time_spent_seconds == 60
    assert worklogs[0].author and worklogs[0].author.account_id == "u1"
    assert worklogs[1].author is None



from __future__ import annotations

from typing import Iterable, List

from ...canonical_models import (
    CanonicalProjectWithOpsgenieTeams,
    JiraProject,
    OpsgenieTeamRef,
)
from ..gen.jira_projects_api import JiraProjectNode, OpsgenieTeamNode


def map_project_with_opsgenie_teams(
    *,
    cloud_id: str,
    project: JiraProjectNode,
    opsgenie_teams: Iterable[OpsgenieTeamNode],
) -> CanonicalProjectWithOpsgenieTeams:
    cloud_id_clean = (cloud_id or "").strip()
    if not cloud_id_clean:
        raise ValueError("cloud_id is required")

    project_key = (project.key or "").strip()
    if not project_key:
        raise ValueError("project.key is required")

    project_name = (project.name or "").strip()
    if not project_name:
        raise ValueError("project.name is required")

    seen: set[str] = set()
    teams: List[OpsgenieTeamRef] = []
    for t in opsgenie_teams:
        team_id = (t.id or "").strip()
        if not team_id:
            raise ValueError("opsgenie_team.id is required")
        if team_id in seen:
            continue
        seen.add(team_id)

        team_name = (t.name or "").strip()
        if not team_name:
            raise ValueError("opsgenie_team.name is required")
        teams.append(OpsgenieTeamRef(id=team_id, name=team_name))

    return CanonicalProjectWithOpsgenieTeams(
        project=JiraProject(cloud_id=cloud_id_clean, key=project_key, name=project_name),
        opsgenie_teams=teams,
    )



import json

import httpx

from atlassian.auth import OAuthBearerAuth
from atlassian.graph.api.jira_sprints import get_sprint_by_id
from atlassian.graph.client import GraphQLClient


def test_jira_sprint_by_id_graphql_mapping():
    def handler(request: httpx.Request) -> httpx.Response:
        payload = json.loads(request.content.decode("utf-8"))
        assert payload.get("operationName") == "JiraSprintById"
        assert payload.get("variables") == {"id": "sprint-1"}
        return httpx.Response(
            200,
            json={
                "data": {
                    "sprintById": {
                        "sprintId": "42",
                        "name": "Sprint 42",
                        "state": "ACTIVE",
                        "startDate": "2021-01-01T00:00:00Z",
                        "endDate": "2021-01-14T00:00:00Z",
                        "completionDate": "2021-01-15T00:00:00Z",
                    }
                }
            },
            request=request,
        )

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://api.atlassian.com",
            auth=OAuthBearerAuth(lambda: "token"),
            http_client=http_client,
        )
        sprint = get_sprint_by_id(client, "sprint-1")

    assert sprint.id == "42"
    assert sprint.name == "Sprint 42"
    assert sprint.state == "ACTIVE"
    assert sprint.start_at == "2021-01-01T00:00:00Z"



from __future__ import annotations

import json
import os
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple


def _add_project_to_syspath() -> None:
    project_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(project_root))


_add_project_to_syspath()

from atlassian.auth import (  # noqa: E402
    BasicApiTokenAuth,
    CookieAuth,
    OAuthBearerAuth,
)
from atlassian.oauth_3lo import OAuthRefreshTokenAuth  # noqa: E402
from atlassian.graph.schema_fetcher import fetch_schema_introspection  # noqa: E402


@dataclass(frozen=True)
class _Config:
    issue_type_name: str
    user_type_name: str


def _env_experimental_apis() -> List[str]:
    raw = os.getenv("ATLASSIAN_GQL_EXPERIMENTAL_APIS", "")
    return [part.strip() for part in raw.split(",") if part.strip()]


def _maybe_strip_quotes(raw: str) -> str:
    if len(raw) >= 2 and raw[0] == raw[-1] and raw[0] in {"'", '"'}:
        return raw[1:-1]
    return raw


def _load_env_file(path: Path) -> None:
    if not path.exists():
        return
    for line in path.read_text(encoding="utf-8").splitlines():
        stripped = line.strip()
        if not stripped or stripped.startswith("#"):
            continue
        if stripped.startswith("export "):
            stripped = stripped[len("export ") :].strip()
        if "=" not in stripped:
            continue
        key, _, value = stripped.partition("=")
        key = key.strip()
        if not key or key in os.environ:
            continue
        os.environ[key] = _maybe_strip_quotes(value.strip())


def _build_auth_from_env():
    token = os.getenv("ATLASSIAN_OAUTH_ACCESS_TOKEN")
    refresh_token = os.getenv("ATLASSIAN_OAUTH_REFRESH_TOKEN")
    client_id = os.getenv("ATLASSIAN_CLIENT_ID")
    client_secret = os.getenv("ATLASSIAN_CLIENT_SECRET")
    email = os.getenv("ATLASSIAN_EMAIL")
    api_token = os.getenv("ATLASSIAN_API_TOKEN")
    cookies_json = os.getenv("ATLASSIAN_COOKIES_JSON")

    if refresh_token and client_id and client_secret:
        return OAuthRefreshTokenAuth(
            client_id=client_id,
            client_secret=client_secret,
            refresh_token=refresh_token,
        )
    if token:
        if client_secret and token.strip() == client_secret.strip():
            raise ValueError(
                "ATLASSIAN_OAUTH_ACCESS_TOKEN appears to be set to ATLASSIAN_CLIENT_SECRET; "
                "set an OAuth access token (not the client secret)."
            )
        return OAuthBearerAuth(lambda: token)
    if email and api_token:
        return BasicApiTokenAuth(email, api_token)
    if cookies_json:
        try:
            cookies = json.loads(cookies_json)
        except json.JSONDecodeError:
            return None
        if isinstance(cookies, dict) and all(
            isinstance(k, str) and isinstance(v, str) for k, v in cookies.items()
        ):
            return CookieAuth(cookies)
    return None


def _load_introspection(path: Path) -> Dict[str, Any]:
    raw = json.loads(path.read_text(encoding="utf-8"))
    if isinstance(raw, dict) and "data" in raw and isinstance(raw["data"], dict):
        data = raw["data"]
    else:
        data = raw
    if not isinstance(data, dict) or "__schema" not in data:
        raise RuntimeError("Introspection JSON missing data.__schema")
    schema = data["__schema"]
    if not isinstance(schema, dict):
        raise RuntimeError("Introspection JSON data.__schema is not an object")
    return schema


def _types_map(schema: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
    types = schema.get("types")
    if not isinstance(types, list):
        raise RuntimeError("Introspection JSON missing __schema.types[]")
    out: Dict[str, Dict[str, Any]] = {}
    for t in types:
        if not isinstance(t, dict):
            continue
        name = t.get("name")
        if isinstance(name, str) and name:
            out[name] = t
    return out


def _unwrap_named_type(type_ref: Dict[str, Any]) -> Tuple[Optional[str], Optional[str], Dict[str, Any]]:
    cur = type_ref
    for _ in range(16):
        if not isinstance(cur, dict):
            break
        kind = cur.get("kind")
        name = cur.get("name")
        if isinstance(name, str) and name:
            return name, kind if isinstance(kind, str) else None, cur
        nxt = cur.get("ofType")
        if nxt is None:
            break
        cur = nxt
    return None, None, {}


def _field(type_def: Dict[str, Any], name: str) -> Optional[Dict[str, Any]]:
    fields = type_def.get("fields")
    if not isinstance(fields, list):
        return None
    for f in fields:
        if isinstance(f, dict) and f.get("name") == name:
            return f
    return None


def _arg(field_def: Dict[str, Any], name: str) -> Optional[Dict[str, Any]]:
    args = field_def.get("args")
    if not isinstance(args, list):
        return None
    for a in args:
        if isinstance(a, dict) and a.get("name") == name:
            return a
    return None


def _require_field(type_def: Dict[str, Any], name: str, path: str) -> Dict[str, Any]:
    found = _field(type_def, name)
    if not found:
        raise RuntimeError(f"Missing required field {path}.{name}")
    return found


def _require_type(types: Dict[str, Dict[str, Any]], name: str) -> Dict[str, Any]:
    type_def = types.get(name)
    if not type_def:
        raise RuntimeError(f"Missing type definition: {name}")
    return type_def


def _discover_config(schema: Dict[str, Any]) -> _Config:
    types = _types_map(schema)

    query_type = schema.get("queryType")
    query_name = query_type.get("name") if isinstance(query_type, dict) else None
    if not isinstance(query_name, str) or not query_name:
        raise RuntimeError("Introspection JSON missing __schema.queryType.name")
    query_def = _require_type(types, query_name)

    issue_by_key_field = _field(query_def, "issueByKey")
    if not issue_by_key_field:
        raise RuntimeError(f"Missing required field {query_name}.issueByKey")
    for arg_name in ("key", "cloudId"):
        if not _arg(issue_by_key_field, arg_name):
            raise RuntimeError(f"Missing required arg issueByKey.{arg_name}")

    issue_type_name, _, _ = _unwrap_named_type(issue_by_key_field.get("type") or {})
    if not issue_type_name:
        raise RuntimeError("Unable to resolve type of issueByKey")
    issue_def = _require_type(types, issue_type_name)

    issue_type_field = _require_field(issue_def, "issueType", f"type {issue_type_name}.fields")
    issue_type_def = _require_type(types, _unwrap_named_type(issue_type_field.get("type") or {})[0] or "")
    _require_field(issue_type_def, "name", f"type {issue_type_def.get('name')}.fields")

    status_field = _require_field(issue_def, "status", f"type {issue_type_name}.fields")
    status_def = _require_type(types, _unwrap_named_type(status_field.get("type") or {})[0] or "")
    _require_field(status_def, "name", f"type {status_def.get('name')}.fields")

    project_field = _require_field(issue_def, "projectField", f"type {issue_type_name}.fields")
    project_field_def = _require_type(types, _unwrap_named_type(project_field.get("type") or {})[0] or "")
    project_field_project = _require_field(project_field_def, "project", f"type {project_field_def.get('name')}.fields")
    project_def = _require_type(types, _unwrap_named_type(project_field_project.get("type") or {})[0] or "")
    _require_field(project_def, "key", f"type {project_def.get('name')}.fields")
    _require_field(project_def, "cloudId", f"type {project_def.get('name')}.fields")

    for dt_field in ("createdField", "updatedField", "resolutionDateField"):
        dt_def = _require_field(issue_def, dt_field, f"type {issue_type_name}.fields")
        dt_type = _require_type(types, _unwrap_named_type(dt_def.get("type") or {})[0] or "")
        _require_field(dt_type, "dateTime", f"type {dt_type.get('name')}.fields")

    assignee_field = _require_field(issue_def, "assigneeField", f"type {issue_type_name}.fields")
    assignee_def = _require_type(types, _unwrap_named_type(assignee_field.get("type") or {})[0] or "")
    _require_field(assignee_def, "user", f"type {assignee_def.get('name')}.fields")

    reporter_field = _require_field(issue_def, "reporter", f"type {issue_type_name}.fields")
    user_type_name, _, _ = _unwrap_named_type(reporter_field.get("type") or {})
    if not user_type_name:
        raise RuntimeError("Unable to resolve reporter user type")

    user_def = _require_type(types, user_type_name)
    _require_field(user_def, "accountId", f"type {user_type_name}.fields")
    _require_field(user_def, "name", f"type {user_type_name}.fields")

    return _Config(issue_type_name=issue_type_name, user_type_name=user_type_name)


def _render_python(cfg: _Config) -> str:
    return f"""# Code generated by python/tools/generate_jira_issue_models.py. DO NOT EDIT.
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional

from atlassian.errors import SerializationError

# Schema types discovered during generation
ISSUE_TYPE_NAME = "{cfg.issue_type_name}"
USER_TYPE_NAME = "{cfg.user_type_name}"

JIRA_ISSUE_BY_KEY_QUERY = \"\"\"query JiraIssueByKey(
  $cloudId: ID!,
  $key: String!
) {{
  issueByKey(key: $key, cloudId: $cloudId) {{
    key
    issueType {{ name }}
    status {{ name }}
    projectField {{
      project {{ key cloudId }}
    }}
    createdField {{ dateTime }}
    updatedField {{ dateTime }}
    resolutionDateField {{ dateTime }}
    assigneeField {{
      user {{ accountId name }}
    }}
    reporter {{ accountId name }}
  }}
}}
\"\"\"


def _expect_dict(obj: Any, path: str) -> Dict[str, Any]:
    if not isinstance(obj, dict):
        raise SerializationError(f\"Expected object at {path}\")
    return obj


def _expect_str(obj: Any, path: str) -> str:
    if not isinstance(obj, str):
        raise SerializationError(f\"Expected string at {path}\")
    return obj


def _expect_optional_str(obj: Any, path: str) -> Optional[str]:
    if obj is None:
        return None
    if not isinstance(obj, str):
        raise SerializationError(f\"Expected string at {path}\")
    if not obj:
        raise SerializationError(f\"Expected non-empty string at {path}\")
    return obj


@dataclass(frozen=True)
class JiraUser:
    account_id: str
    name: str

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"JiraUser\":
        raw = _expect_dict(obj, path)
        return JiraUser(
            account_id=_expect_str(raw.get(\"accountId\"), f\"{path}.accountId\"),
            name=_expect_str(raw.get(\"name\"), f\"{path}.name\"),
        )


@dataclass(frozen=True)
class JiraIssueType:
    name: str

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"JiraIssueType\":
        raw = _expect_dict(obj, path)
        return JiraIssueType(name=_expect_str(raw.get(\"name\"), f\"{path}.name\"))


@dataclass(frozen=True)
class JiraStatus:
    name: str

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"JiraStatus\":
        raw = _expect_dict(obj, path)
        return JiraStatus(name=_expect_str(raw.get(\"name\"), f\"{path}.name\"))


@dataclass(frozen=True)
class JiraProject:
    key: str
    cloud_id: str

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"JiraProject\":
        raw = _expect_dict(obj, path)
        return JiraProject(
            key=_expect_str(raw.get(\"key\"), f\"{path}.key\"),
            cloud_id=_expect_str(raw.get(\"cloudId\"), f\"{path}.cloudId\"),
        )


@dataclass(frozen=True)
class JiraProjectField:
    project: JiraProject

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"JiraProjectField\":
        raw = _expect_dict(obj, path)
        return JiraProjectField(
            project=JiraProject.from_dict(raw.get(\"project\"), f\"{path}.project\"),
        )


@dataclass(frozen=True)
class JiraDateTimePickerField:
    date_time: Optional[str]

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"JiraDateTimePickerField\":
        raw = _expect_dict(obj, path)
        return JiraDateTimePickerField(
            date_time=_expect_optional_str(raw.get(\"dateTime\"), f\"{path}.dateTime\")
        )


@dataclass(frozen=True)
class JiraSingleSelectUserPickerField:
    user: Optional[JiraUser]

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"JiraSingleSelectUserPickerField\":
        raw = _expect_dict(obj, path)
        user_raw = raw.get(\"user\")
        user = JiraUser.from_dict(user_raw, f\"{path}.user\") if user_raw is not None else None
        return JiraSingleSelectUserPickerField(user=user)


@dataclass(frozen=True)
class JiraIssueNode:
    key: str
    issue_type: JiraIssueType
    status: JiraStatus
    project_field: JiraProjectField
    created_field: JiraDateTimePickerField
    updated_field: JiraDateTimePickerField
    resolution_date_field: Optional[JiraDateTimePickerField]
    assignee_field: Optional[JiraSingleSelectUserPickerField]
    reporter: Optional[JiraUser]

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"JiraIssueNode\":
        raw = _expect_dict(obj, path)
        resolution_raw = raw.get(\"resolutionDateField\")
        assignee_raw = raw.get(\"assigneeField\")
        reporter_raw = raw.get(\"reporter\")
        return JiraIssueNode(
            key=_expect_str(raw.get(\"key\"), f\"{path}.key\"),
            issue_type=JiraIssueType.from_dict(raw.get(\"issueType\"), f\"{path}.issueType\"),
            status=JiraStatus.from_dict(raw.get(\"status\"), f\"{path}.status\"),
            project_field=JiraProjectField.from_dict(raw.get(\"projectField\"), f\"{path}.projectField\"),
            created_field=JiraDateTimePickerField.from_dict(raw.get(\"createdField\"), f\"{path}.createdField\"),
            updated_field=JiraDateTimePickerField.from_dict(raw.get(\"updatedField\"), f\"{path}.updatedField\"),
            resolution_date_field=JiraDateTimePickerField.from_dict(resolution_raw, f\"{path}.resolutionDateField\")
            if resolution_raw is not None
            else None,
            assignee_field=JiraSingleSelectUserPickerField.from_dict(assignee_raw, f\"{path}.assigneeField\")
            if assignee_raw is not None
            else None,
            reporter=JiraUser.from_dict(reporter_raw, f\"{path}.reporter\") if reporter_raw is not None else None,
        )


def parse_jira_issue_by_key(data: Any) -> JiraIssueNode:
    root = _expect_dict(data, \"data\")
    issue = root.get(\"issueByKey\")
    if issue is None:
        raise SerializationError(\"Missing data.issueByKey\")
    return JiraIssueNode.from_dict(issue, \"data.issueByKey\")
"""


def main() -> None:
    repo_root = Path(__file__).resolve().parents[1]
    token_file = Path(os.getenv("ATLASSIAN_OAUTH_TOKEN_FILE", repo_root / "oauth_tokens.txt"))
    _load_env_file(token_file)

    schema_path = repo_root / "graphql" / "schema.introspection.json"
    if not schema_path.exists():
        base_url = os.getenv("ATLASSIAN_GQL_BASE_URL")
        if not base_url and (
            os.getenv("ATLASSIAN_OAUTH_ACCESS_TOKEN") or os.getenv("ATLASSIAN_OAUTH_REFRESH_TOKEN")
        ):
            base_url = "https://api.atlassian.com"
        if not base_url:
            raise RuntimeError(f"Missing {schema_path} and ATLASSIAN_GQL_BASE_URL not set")
        auth = _build_auth_from_env()
        if auth is None:
            raise RuntimeError("No credentials available in env vars to fetch schema")
        fetch_schema_introspection(
            base_url=base_url,
            auth=auth,
            output_dir=str(schema_path.parent),
            experimental_apis=_env_experimental_apis(),
        )

    schema = _load_introspection(schema_path)
    cfg = _discover_config(schema)

    out_path = repo_root / "python" / "atlassian" / "graph" / "gen" / "jira_issues_api.py"
    out_path.parent.mkdir(parents=True, exist_ok=True)
    out_path.write_text(_render_python(cfg), encoding="utf-8")
    print(f"Wrote {out_path}")


if __name__ == "__main__":
    main()



from __future__ import annotations

from ...canonical_models import JiraBoard
from ..gen import jira_agile_api as api


def map_rest_board(board: api.Board) -> JiraBoard:
    return JiraBoard(
        id=str(board.id) if board.id is not None else "",
        name=board.name or "",
        type=board.board_type or "",
    )



import json

import httpx

from atlassian.graph.client import GraphQLClient
from atlassian.oauth_3lo import OAuthRefreshTokenAuth


def test_oauth_refresh_token_auth_applies_and_caches_token():
    calls = {"token": 0, "graphql": 0, "auth": []}

    def handler(request: httpx.Request) -> httpx.Response:
        if request.url.path == "/oauth/token":
            calls["token"] += 1
            payload = json.loads(request.content.decode("utf-8"))
            assert payload["grant_type"] == "refresh_token"
            assert payload["client_id"] == "client-id"
            assert payload["client_secret"] == "client-secret"
            assert payload["refresh_token"] == "refresh-token"
            return httpx.Response(
                200,
                json={
                    "access_token": "access-1",
                    "token_type": "Bearer",
                    "expires_in": 3600,
                },
                request=request,
            )
        if request.url.path == "/graphql":
            calls["graphql"] += 1
            calls["auth"].append(request.headers.get("Authorization"))
            return httpx.Response(200, json={"data": {"ok": True}}, request=request)
        return httpx.Response(404, json={"error": "not found"}, request=request)

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        auth = OAuthRefreshTokenAuth(
            client_id="client-id",
            client_secret="client-secret",
            refresh_token="refresh-token",
            token_url="https://example.com/oauth/token",
            http_client=http_client,
        )
        client = GraphQLClient(
            "https://example.com",
            auth=auth,
            http_client=http_client,
        )
        client.execute("query { ok }")
        client.execute("query { ok }")

    assert calls["token"] == 1
    assert calls["graphql"] == 2
    assert calls["auth"] == ["Bearer access-1", "Bearer access-1"]



import json
import logging
import os
from pathlib import Path

import pytest

from atlassian import (
    BasicApiTokenAuth,
    CookieAuth,
    JiraRestClient,
    OAuthBearerAuth,
    OAuthRefreshTokenAuth,
    RateLimitError,
)
from atlassian.rest.api.jira_projects import iter_projects_via_rest


def _load_dotenv_if_present() -> None:
    env_path = Path(__file__).resolve().parents[3] / ".env"
    if not env_path.exists():
        return
    for raw in env_path.read_text(encoding="utf-8").splitlines():
        line = raw.strip()
        if not line or line.startswith("#"):
            continue
        if line.startswith("export "):
            line = line[len("export ") :].strip()
        if "=" not in line:
            continue
        key, value = line.split("=", 1)
        key = key.strip()
        if not key or key in os.environ:
            continue
        value = value.strip()
        if (value.startswith('"') and value.endswith('"')) or (value.startswith("'") and value.endswith("'")):
            value = value[1:-1]
        os.environ[key] = value


def _get_auth():
    token = os.getenv("ATLASSIAN_OAUTH_ACCESS_TOKEN")
    refresh_token = os.getenv("ATLASSIAN_OAUTH_REFRESH_TOKEN")
    client_id = os.getenv("ATLASSIAN_CLIENT_ID")
    client_secret = os.getenv("ATLASSIAN_CLIENT_SECRET")
    email = os.getenv("ATLASSIAN_EMAIL")
    api_token = os.getenv("ATLASSIAN_API_TOKEN")
    cookies_json = os.getenv("ATLASSIAN_COOKIES_JSON")

    if refresh_token and client_id and client_secret:
        return OAuthRefreshTokenAuth(
            client_id=client_id,
            client_secret=client_secret,
            refresh_token=refresh_token,
        )
    if token:
        if client_secret and token.strip() == client_secret.strip():
            pytest.fail(
                "ATLASSIAN_OAUTH_ACCESS_TOKEN appears to be set to ATLASSIAN_CLIENT_SECRET; "
                "set an actual OAuth access token (not the client secret)."
            )
        return OAuthBearerAuth(lambda: token)
    if email and api_token:
        return BasicApiTokenAuth(email, api_token)
    if cookies_json:
        try:
            cookies = json.loads(cookies_json)
            if isinstance(cookies, dict):
                return CookieAuth(cookies)
        except json.JSONDecodeError:
            pass
    return None


def _derive_site_base_url_from_gql_base(gql_base_url: str):
    candidate = (gql_base_url or "").strip().rstrip("/")
    if not candidate:
        return None
    for suffix in ("/gateway/api/graphql", "/gateway/api", "/graphql"):
        if candidate.endswith(suffix):
            return candidate[: -len(suffix)].rstrip("/")
    return None


def _jira_base_url(auth, cloud_id: str):
    explicit = os.getenv("ATLASSIAN_JIRA_BASE_URL")
    if explicit:
        return explicit.strip().rstrip("/")

    if isinstance(auth, (OAuthBearerAuth, OAuthRefreshTokenAuth)):
        return f"https://api.atlassian.com/ex/jira/{cloud_id}"

    gql_base_url = os.getenv("ATLASSIAN_GQL_BASE_URL")
    if gql_base_url:
        return _derive_site_base_url_from_gql_base(gql_base_url)
    return None


def _cloud_id():
    return os.getenv("ATLASSIAN_CLOUD_ID") or os.getenv("ATLASSIAN_JIRA_CLOUD_ID")


def test_live_jira_projects_rest_smoke():
    _load_dotenv_if_present()
    auth = _get_auth()
    if auth is None:
        pytest.skip("Integration credentials not provided")

    cloud_id = _cloud_id()
    if not cloud_id:
        pytest.fail(
            "ATLASSIAN_CLOUD_ID (or ATLASSIAN_JIRA_CLOUD_ID) is required when running Jira REST integration tests"
        )

    base_url = _jira_base_url(auth, cloud_id)
    if not base_url:
        pytest.skip(
            "ATLASSIAN_JIRA_BASE_URL not set and could not derive Jira base URL "
            "(set ATLASSIAN_JIRA_BASE_URL or ATLASSIAN_GQL_BASE_URL for tenanted auth)"
        )

    logger = logging.getLogger("atlassian.integration")
    client = JiraRestClient(base_url, auth=auth, timeout_seconds=30.0, logger=logger, max_retries_429=1)
    try:
        it = iter_projects_via_rest(
            client,
            cloud_id=cloud_id,
            project_types=["SOFTWARE"],
            page_size=50,
        )
        try:
            first = next(it, None)
        except RateLimitError as exc:
            pytest.skip(f"Rate limited during integration; {exc}")
        if first is not None:
            assert first.project.cloud_id == cloud_id
            assert first.project.key
            assert first.project.name
            assert first.project.type == "SOFTWARE"
            assert first.opsgenie_teams == []
    finally:
        client.close()



from datetime import datetime, timedelta, timezone

import httpx
import pytest

from atlassian.auth import OAuthBearerAuth
from atlassian.errors import (
    GraphQLOperationError,
    LocalRateLimitError,
    RateLimitError,
    SerializationError,
    TransportError,
)
from atlassian.graph.client import GraphQLClient


def test_execute_returns_data():
    def handler(request: httpx.Request):
        return httpx.Response(200, json={"data": {"ok": True}}, request=request)

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://api.atlassian.com",
            auth=OAuthBearerAuth(lambda: "token"),
            http_client=http_client,
        )
        result = client.execute("query { ok }")
        assert result.data == {"ok": True}


def test_strict_mode_raises_on_errors():
    def handler(request: httpx.Request):
        return httpx.Response(
            200,
            json={"data": {"partial": True}, "errors": [{"message": "bad"}]},
            request=request,
        )

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://api.atlassian.com",
            auth=OAuthBearerAuth(lambda: "token"),
            strict=True,
            http_client=http_client,
        )
        with pytest.raises(GraphQLOperationError):
            client.execute("query { broken }")


def test_retries_on_429_with_retry_after_timestamp():
    now = datetime(2021, 5, 10, 10, 59, 58, tzinfo=timezone.utc)
    current = {"now": now}

    def now_fn():
        return current["now"]

    slept: list[float] = []

    def sleeper(seconds: float) -> None:
        slept.append(seconds)
        current["now"] = current["now"] + timedelta(seconds=seconds)

    responses = [
        lambda request: httpx.Response(
            429,
            headers={"Retry-After": "2021-05-10T11:00Z"},
            request=request,
        ),
        lambda request: httpx.Response(200, json={"data": {"ok": True}}, request=request),
    ]

    def handler(request: httpx.Request):
        return responses.pop(0)(request)

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://api.atlassian.com",
            auth=OAuthBearerAuth(lambda: "token"),
            max_retries_429=1,
            time_provider=now_fn,
            sleeper=sleeper,
            http_client=http_client,
        )
        result = client.execute("query { ok }")
        assert result.data == {"ok": True}
        assert slept and pytest.approx(slept[0], rel=0.01) == 2.0


def test_raises_rate_limit_error_on_invalid_retry_after():
    def handler(request: httpx.Request):
        return httpx.Response(
            429,
            headers={"Retry-After": "invalid"},
            request=request,
        )

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://api.atlassian.com",
            auth=OAuthBearerAuth(lambda: "token"),
            max_retries_429=0,
            http_client=http_client,
        )
        with pytest.raises(RateLimitError) as excinfo:
            client.execute("query { ok }")
        assert excinfo.value.header_value == "invalid"
        assert excinfo.value.attempts == 1


def test_retry_after_in_past_retries_immediately():
    now = datetime(2021, 5, 10, 11, 0, 1, tzinfo=timezone.utc)
    current = {"now": now}

    def now_fn():
        return current["now"]

    slept: list[float] = []

    def sleeper(seconds: float) -> None:
        slept.append(seconds)
        current["now"] = current["now"] + timedelta(seconds=seconds)

    responses = [
        lambda request: httpx.Response(
            429,
            headers={"Retry-After": "2021-05-10T11:00Z"},
            request=request,
        ),
        lambda request: httpx.Response(200, json={"data": {"ok": True}}, request=request),
    ]

    def handler(request: httpx.Request):
        return responses.pop(0)(request)

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://api.atlassian.com",
            auth=OAuthBearerAuth(lambda: "token"),
            max_retries_429=1,
            time_provider=now_fn,
            sleeper=sleeper,
            http_client=http_client,
        )
        result = client.execute("query { ok }")
        assert result.data == {"ok": True}
        assert slept == []


@pytest.mark.parametrize("status_code", [500, 502, 503])
def test_does_not_retry_on_5xx(status_code):
    call_count = {"count": 0}

    def handler(request):
        call_count["count"] += 1
        return httpx.Response(status_code, request=request)

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://api.atlassian.com",
            auth=OAuthBearerAuth(lambda: "token"),
            max_retries_429=2,
            http_client=http_client,
        )
        with pytest.raises(TransportError):
            client.execute("query { fail }")
        assert call_count["count"] == 1


def test_local_throttling_fails_when_wait_exceeds_max():
    now = datetime(2021, 5, 10, 10, 0, 0, tzinfo=timezone.utc)
    current = {"now": now}

    def now_fn():
        return current["now"]

    slept: list[float] = []

    def sleeper(seconds: float) -> None:
        slept.append(seconds)
        current["now"] = current["now"] + timedelta(seconds=seconds)

    call_count = {"count": 0}

    def handler(request: httpx.Request):
        call_count["count"] += 1
        return httpx.Response(200, json={"data": {"ok": True}}, request=request)

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://api.atlassian.com",
            auth=OAuthBearerAuth(lambda: "token"),
            enable_local_throttling=True,
            max_wait_seconds=5,
            time_provider=now_fn,
            sleeper=sleeper,
            http_client=http_client,
        )
        with pytest.raises(LocalRateLimitError):
            client.execute("query { ok }", estimated_cost=20000)
        assert call_count["count"] == 0
        assert slept  # waited locally before giving up


def test_invalid_json_raises_serialization_error():
    def handler(request: httpx.Request):
        return httpx.Response(200, content=b"not-json", request=request)

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://api.atlassian.com",
            auth=OAuthBearerAuth(lambda: "token"),
            http_client=http_client,
        )
        with pytest.raises(SerializationError):
            client.execute("query { ok }")
